//
// libsyd: Rust-based C library for SydB☮x interaction via /dev/syd
// lib/src/lib.rs: SydB☮x API C Library
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: LGPL-3.0-or-later

//! # libsyd - SydB☮x API Rust Library
//!
//! `libsyd` is a C library written in Rust that implements the SydB☮x
//! stat API, providing an interface to the `/dev/syd` of Sydb☮x. It
//! allows for runtime configuration and interaction with the Sydb☮x
//! sandboxing environment.
//!
//! ## Overview
//! The library is designed to interact with the Sydb☮x sandboxing
//! environment, offering functionalities to check and modify the state
//! of the sandbox lock, and perform system calls to `/dev/syd`.
//!
//! ## Attention
//! This library is a work in progress, and **the API is not stable**.
//!
//! For more detailed information and usage instructions, refer to the Sydb☮x
//! manual, available at [Sydb☮x Manual](https://git.sr.ht/~alip/syd).
//!
//! ## Author
//! Ali Polatel <alip@chesswob.org>

// We like safe, clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    ffi::{CStr, CString, OsStr, OsString},
    os::{
        raw::{c_char, c_int, c_uint},
        unix::ffi::{OsStrExt, OsStringExt},
    },
    path::{Path, PathBuf},
};

use nix::{errno::Errno, NixPath};

/// An enumeration of the possible states for the sandbox lock.
#[repr(C)]
pub enum LockState {
    /// The sandbox lock is off, allowing all sandbox commands.
    Off,
    /// The sandbox lock is set to on for all processes except the initial
    /// process (Sydb☮x exec child). This is the default state.
    Exec,
    /// The sandbox lock is on, disallowing all sandbox commands.
    On,
}

include!("syd.rs");

/// Performs a Sydb☮x API check
///
/// The caller is advised to perform this check before
/// calling any other Sydb☮x API calls.
///
/// Returns API number on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_api() -> c_int {
    match stat("/dev/syd/3") {
        0 => 3,
        n => n,
    }
}

/// Performs an lstat system call on the file "/dev/syd".
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_check() -> c_int {
    stat("/dev/syd")
}

/// Causes Sydb☮x to exit immediately with code 127
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_panic() -> c_int {
    stat("/dev/syd/panic")
}

/// Causes Sydb☮x to reset sandboxing to the default state.
/// Allowlists, denylists and filters are going to be cleared.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_reset() -> c_int {
    stat("/dev/syd/reset")
}

/// Causes Sydb☮x to read configuration from the given file descriptor.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_load(fd: c_int) -> c_int {
    let path = PathBuf::from("/dev/syd/load");

    // Convert the file descriptor directly to ASCII bytes
    let fd = if fd == 0 {
        vec![b'0'] // Handle zero separately
    } else {
        let mut n = fd;
        let mut bytes = Vec::new();
        while n > 0 {
            bytes.insert(0, (n % 10) as u8 + b'0'); // Convert each digit to ASCII
            n /= 10;
        }
        bytes
    };

    // Create an OsString from the raw byte vector
    stat(path.join(OsString::from_vec(fd)))
}

/// Sets the state of the sandbox lock.
///
/// state: The desired state of the sandbox lock.
///
/// Returns Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_lock(state: c_uint) -> c_int {
    match state {
        0 => stat("/dev/syd/lock:off"),
        1 => stat("/dev/syd/lock:exec"),
        2 => stat("/dev/syd/lock:on"),
        _ => -EINVAL,
    }
}

/// Checks if memory sandboxing is enabled.
///
/// Returns true if memory sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_mem() -> bool {
    stat("/dev/syd/sandbox/mem?") == 0
}

/// Enable memory sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_mem() -> c_int {
    stat("/dev/syd/sandbox/mem:on")
}

/// Disable memory sandboxing.
///
/// Returns Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_mem() -> c_int {
    stat("/dev/syd/sandbox/mem:off")
}

/// Checks if pid sandboxing is enabled.
///
/// Returns true if pid sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_pid() -> bool {
    stat("/dev/syd/sandbox/pid?") == 0
}

/// Enable pid sandboxing.
///
/// Returns Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_pid() -> c_int {
    stat("/dev/syd/sandbox/pid:on")
}

/// Disable pid sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_pid() -> c_int {
    stat("/dev/syd/sandbox/pid:off")
}

/// Checks if read sandboxing is enabled.
///
/// Returns true if read sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_read() -> bool {
    stat("/dev/syd/sandbox/read?") == 0
}

/// Enable read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_read() -> c_int {
    stat("/dev/syd/sandbox/read:on")
}

/// Disable read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_read() -> c_int {
    stat("/dev/syd/sandbox/read:off")
}

/// Checks if stat sandboxing is enabled.
///
/// Returns true if stat sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_stat() -> bool {
    stat("/dev/syd/sandbox/stat?") == 0
}

/// Enable stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_stat() -> c_int {
    stat("/dev/syd/sandbox/stat:on")
}

/// Disable stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_stat() -> c_int {
    stat("/dev/syd/sandbox/stat:off")
}

/// Checks if write sandboxing is enabled.
///
/// Returns true if write sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_write() -> bool {
    stat("/dev/syd/sandbox/write?") == 0
}

/// Enable write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_write() -> c_int {
    stat("/dev/syd/sandbox/write:on")
}

/// Disable write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_write() -> c_int {
    stat("/dev/syd/sandbox/write:off")
}

/// Checks if exec sandboxing is enabled.
///
/// Returns true if exec sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_exec() -> bool {
    stat("/dev/syd/sandbox/exec?") == 0
}

/// Enable exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_exec() -> c_int {
    stat("/dev/syd/sandbox/exec:on")
}

/// Disable exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_exec() -> c_int {
    stat("/dev/syd/sandbox/exec:off")
}

/// Checks if net sandboxing is enabled.
///
/// Returns true if net sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_net() -> bool {
    stat("/dev/syd/sandbox/net?") == 0
}

/// Enable net sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_net() -> c_int {
    stat("/dev/syd/sandbox/net:on")
}

/// Disable net sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_net() -> c_int {
    stat("/dev/syd/sandbox/net:off")
}

/// Adds to the list of glob patterns used to determine which paths
/// should be killed (prevented from executing) in the sandbox.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_kill_add(glob: *const c_char) -> c_int {
    esyd("exec/kill", glob, b'+')
}

/// Deletes the first matching item from the end of the list of glob
/// patterns used to determine which paths should be killed (prevented
/// from executing) in the sandbox.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_kill_del(glob: *const c_char) -> c_int {
    esyd("exec/kill", glob, b'-')
}

/// Removes all matching items from the list of glob patterns used to
/// determine which paths should be killed (prevented from executing) in
/// the sandbox.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_kill_rem(glob: *const c_char) -> c_int {
    esyd("exec/kill", glob, b'^')
}

/// Adds to the allowlist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_read_add(glob: *const c_char) -> c_int {
    esyd("allowlist/read", glob, b'+')
}

/// Removes the first instance from the end of the allowlist of read
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_read_del(glob: *const c_char) -> c_int {
    esyd("allowlist/read", glob, b'-')
}

/// Removes all matching patterns from the allowlist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_read_rem(glob: *const c_char) -> c_int {
    esyd("allowlist/read", glob, b'^')
}

/// Adds to the denylist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_read_add(glob: *const c_char) -> c_int {
    esyd("denylist/read", glob, b'+')
}

/// Removes the first instance from the end of the denylist of read
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_read_del(glob: *const c_char) -> c_int {
    esyd("denylist/read", glob, b'-')
}

/// Removes all matching patterns from the denylist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_read_rem(glob: *const c_char) -> c_int {
    esyd("denylist/read", glob, b'^')
}

/// Adds to the filter of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_read_add(glob: *const c_char) -> c_int {
    esyd("filter/read", glob, b'+')
}

/// Removes the first instance from the end of the filter of read
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_read_del(glob: *const c_char) -> c_int {
    esyd("filter/read", glob, b'-')
}

/// Removes all matching patterns from the filter of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_read_rem(glob: *const c_char) -> c_int {
    esyd("filter/read", glob, b'^')
}

/// Adds to the allowlist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_stat_add(glob: *const c_char) -> c_int {
    esyd("allowlist/stat", glob, b'+')
}

/// Removes the first instance from the end of the allowlist of stat
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_stat_del(glob: *const c_char) -> c_int {
    esyd("allowlist/stat", glob, b'-')
}

/// Removes all matching patterns from the allowlist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_stat_rem(glob: *const c_char) -> c_int {
    esyd("allowlist/stat", glob, b'^')
}

/// Adds to the denylist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_stat_add(glob: *const c_char) -> c_int {
    esyd("denylist/stat", glob, b'+')
}

/// Removes the first instance from the end of the denylist of stat
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_stat_del(glob: *const c_char) -> c_int {
    esyd("denylist/stat", glob, b'-')
}

/// Removes all matching patterns from the denylist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_stat_rem(glob: *const c_char) -> c_int {
    esyd("denylist/stat", glob, b'^')
}

/// Adds to the filter of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_stat_add(glob: *const c_char) -> c_int {
    esyd("filter/stat", glob, b'+')
}

/// Removes the first instance from the end of the filter of stat
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_stat_del(glob: *const c_char) -> c_int {
    esyd("filter/stat", glob, b'-')
}

/// Removes all matching patterns from the filter of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_stat_rem(glob: *const c_char) -> c_int {
    esyd("filter/stat", glob, b'^')
}

/// Adds to the allowlist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_write_add(glob: *const c_char) -> c_int {
    esyd("allowlist/write", glob, b'+')
}

/// Removes the first instance from the end of the allowlist of write
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_write_del(glob: *const c_char) -> c_int {
    esyd("allowlist/write", glob, b'-')
}

/// Removes all matching patterns from the allowlist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_write_rem(glob: *const c_char) -> c_int {
    esyd("allowlist/write", glob, b'^')
}

/// Adds to the denylist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_write_add(glob: *const c_char) -> c_int {
    esyd("denylist/write", glob, b'+')
}

/// Removes the first instance from the end of the denylist of write
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_write_del(glob: *const c_char) -> c_int {
    esyd("denylist/write", glob, b'-')
}

/// Removes all matching patterns from the denylist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_write_rem(glob: *const c_char) -> c_int {
    esyd("denylist/write", glob, b'^')
}

/// Adds to the filter of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_write_add(glob: *const c_char) -> c_int {
    esyd("filter/write", glob, b'+')
}

/// Removes the first instance from the end of the filter of write
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_write_del(glob: *const c_char) -> c_int {
    esyd("filter/write", glob, b'-')
}

/// Removes all matching patterns from the filter of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_write_rem(glob: *const c_char) -> c_int {
    esyd("filter/write", glob, b'^')
}

/// Adds to the allowlist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_exec_add(glob: *const c_char) -> c_int {
    esyd("allowlist/exec", glob, b'+')
}

/// Removes the first instance from the end of the allowlist of exec
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_exec_del(glob: *const c_char) -> c_int {
    esyd("allowlist/exec", glob, b'-')
}

/// Removes all matching patterns from the allowlist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_exec_rem(glob: *const c_char) -> c_int {
    esyd("allowlist/exec", glob, b'^')
}

/// Adds to the denylist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_exec_add(glob: *const c_char) -> c_int {
    esyd("denylist/exec", glob, b'+')
}

/// Removes the first instance from the end of the denylist of exec
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_exec_del(glob: *const c_char) -> c_int {
    esyd("denylist/exec", glob, b'-')
}

/// Removes all matching patterns from the denylist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_exec_rem(glob: *const c_char) -> c_int {
    esyd("denylist/exec", glob, b'^')
}

/// Adds to the filter of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_exec_add(glob: *const c_char) -> c_int {
    esyd("filter/exec", glob, b'+')
}

/// Removes the first instance from the end of the filter of exec
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_exec_del(glob: *const c_char) -> c_int {
    esyd("filter/exec", glob, b'-')
}

/// Removes all matching patterns from the filter of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_exec_rem(glob: *const c_char) -> c_int {
    esyd("filter/exec", glob, b'^')
}

/// Adds to the allowlist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_bind_add(addr: *const c_char) -> c_int {
    esyd("allowlist/net/bind", addr, b'+')
}

/// Removes the first instance from the end of the allowlist of net/bind
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_bind_del(addr: *const c_char) -> c_int {
    esyd("allowlist/net/bind", addr, b'-')
}

/// Removes all matching patterns from the allowlist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_bind_rem(addr: *const c_char) -> c_int {
    esyd("allowlist/net/bind", addr, b'^')
}

/// Adds to the denylist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_bind_add(addr: *const c_char) -> c_int {
    esyd("denylist/net/bind", addr, b'+')
}

/// Removes the first instance from the end of the denylist of net/bind
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_bind_del(addr: *const c_char) -> c_int {
    esyd("denylist/net/bind", addr, b'-')
}

/// Removes all matching patterns from the denylist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_bind_rem(addr: *const c_char) -> c_int {
    esyd("denylist/net/bind", addr, b'^')
}

/// Adds to the filter of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_bind_add(addr: *const c_char) -> c_int {
    esyd("filter/net/bind", addr, b'+')
}

/// Removes the first instance from the end of the filter of net/bind
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_bind_del(addr: *const c_char) -> c_int {
    esyd("filter/net/bind", addr, b'-')
}

/// Removes all matching patterns from the filter of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_bind_rem(addr: *const c_char) -> c_int {
    esyd("filter/net/bind", addr, b'^')
}

/// Adds to the allowlist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_connect_add(addr: *const c_char) -> c_int {
    esyd("allowlist/net/connect", addr, b'+')
}

/// Removes the first instance from the end of the allowlist of net/connect
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_connect_del(addr: *const c_char) -> c_int {
    esyd("allowlist/net/connect", addr, b'-')
}

/// Removes all matching patterns from the allowlist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_allow_net_connect_rem(addr: *const c_char) -> c_int {
    esyd("allowlist/net/connect", addr, b'^')
}

/// Adds to the denylist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_connect_add(addr: *const c_char) -> c_int {
    esyd("denylist/net/connect", addr, b'+')
}

/// Removes the first instance from the end of the denylist of net/connect
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_connect_del(addr: *const c_char) -> c_int {
    esyd("denylist/net/connect", addr, b'-')
}

/// Removes all matching patterns from the denylist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_deny_net_connect_rem(addr: *const c_char) -> c_int {
    esyd("denylist/net/connect", addr, b'^')
}

/// Adds to the filter of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_connect_add(addr: *const c_char) -> c_int {
    esyd("filter/net/connect", addr, b'+')
}

/// Removes the first instance from the end of the filter of net/connect
/// sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_connect_del(addr: *const c_char) -> c_int {
    esyd("filter/net/connect", addr, b'-')
}

/// Removes all matching patterns from the filter of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_filter_net_connect_rem(addr: *const c_char) -> c_int {
    esyd("filter/net/connect", addr, b'^')
}

/// Set SydB☮x maximum per-process memory usage limit for memory sandboxing,
/// parse-size crate is used to parse the value so formatted strings are OK.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_mem_max(size: *const c_char) -> c_int {
    esyd("mem/max", size, b':')
}

/// Set SydB☮x maximum per-process virtual memory usage limit for memory sandboxing,
/// parse-size crate is used to parse the value so formatted strings are OK.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_mem_vm_max(size: *const c_char) -> c_int {
    esyd("mem/vm_max", size, b':')
}

/// Set SydB☮x maximum process id limit for pid sandboxing
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_pid_max(size: usize) -> c_int {
    let size = usize_to_cstring(size);
    esyd("pid/max", size.as_ptr(), b':')
}

fn usize_to_cstring(value: usize) -> CString {
    let digits = if value == 0 {
        vec![b'0']
    } else {
        let mut number = value;
        let mut digits = Vec::new();
        while number > 0 {
            digits.push(b'0' + (number % 10) as u8);
            number /= 10;
        }
        digits.reverse();
        digits
    };

    // SAFETY: The use of CString::from_vec_unchecked is safe here because:
    // 1. We only push ASCII digits to the vector, so there are no interior null bytes.
    // 2. The function appends the necessary terminating null byte.
    unsafe { CString::from_vec_unchecked(digits) }
}

/// Execute a command outside the sandbox without sandboxing
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_exec(file: *const c_char, argv: *const *const c_char) -> c_int {
    if file.is_null() || argv.is_null() {
        return -EFAULT;
    }

    // SAFETY: We're creating a slice from `file` with a maximum length
    // of CMDMAX. This is safe under the assumption that `file` points
    // to at least CMDMAX bytes of valid memory.
    let file = unsafe { std::slice::from_raw_parts(file as *const u8, CMDMAX) };
    let file = match CStr::from_bytes_until_nul(file) {
        Ok(file) => OsStr::from_bytes(file.to_bytes()),
        Err(_) => return -ENAMETOOLONG,
    };

    let mut path = OsString::from("/dev/syd/cmd/exec!");
    path.push(file);

    let mut idx: isize = 0;
    while unsafe { !(*argv.offset(idx)).is_null() } {
        // SAFETY: Similar to `file`, we're creating a slice from each
        // `argv` element, limiting the length to CMDMAX. We've checked
        // for null pointers above.
        let arg = unsafe { std::slice::from_raw_parts(*argv.offset(idx) as *const u8, CMDMAX) };
        let arg = match CStr::from_bytes_until_nul(arg) {
            Ok(arg) => OsStr::from_bytes(arg.to_bytes()),
            Err(_) => return -EINVAL,
        };

        path.push(OsStr::from_bytes(&[b'\x1F'])); // ASCII Unit Separator
        path.push(arg);

        // SAFETY: We do not allow more than ARGMAX arguments.
        idx = idx.saturating_add(1);
        if idx >= ARGMAX {
            return -ENAMETOOLONG;
        }
    }

    let path = PathBuf::from(path);
    stat(path)
}
