//
// libsyd: Rust-based C library for SydB☮x interaction via /dev/syd
// lib/src/syd.rs: SydB☮x API C Library, critical public functions
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: LGPL-3.0-or-later

const ARGMAX: isize = 16;
const CMDMAX: usize = 64;

const EFAULT: i32 = 14;
const EINVAL: i32 = 22;
const ENAMETOOLONG: i32 = 36;

fn stat<P: AsRef<Path>>(path: P) -> c_int {
    match path.as_ref().with_nix_path(|cstr| unsafe {
        nix::libc::fstatat(nix::libc::AT_FDCWD,
            cstr.as_ptr(),
            std::ptr::null_mut::<nix::libc::stat>(),
            nix::libc::AT_SYMLINK_NOFOLLOW)
    }) {
        Ok(0) => 0,
        Ok(_) => -(Errno::last() as i32),
        Err(errno) => -(errno as i32),
    }
}

fn esyd<P: AsRef<Path>>(rule: P, elem: *const c_char, op: u8) -> c_int {
    if !matches!(op, b'+' | b'-' | b'^' | b':') {
        return -EINVAL;
    }

    if elem.is_null() {
        return -EFAULT;
    }

    // SAFETY: The `from_raw_parts` function is used to create a byte slice
    // from the `elem` raw pointer. We limit the slice length to CMDMAX to
    // avoid exceeding the bounds of the memory that `elem` might point to.
    // This slice creation is safe under the assumption that `elem` points to
    // at least CMDMAX bytes of valid memory.
    let elem = unsafe { std::slice::from_raw_parts(elem as *const u8, CMDMAX) };

    // Attempt to create a CStr from the byte slice,
    // stopping at the first null byte or at CMDMAX.
    let elem = match CStr::from_bytes_until_nul(elem) {
        Ok(elem) => OsStr::from_bytes(elem.to_bytes()),
        Err(_) => return -ENAMETOOLONG,
    };

    // Manually concatenate the path segments
    let mut path = OsString::from("/dev/syd/");
    path.push(rule.as_ref());
    path.push(OsStr::from_bytes(&[op]));
    path.push(elem);

    // Convert the OsString to PathBuf
    let path = PathBuf::from(path);

    stat(path)
}
